# (Be in -*- python -*- mode.)
#
# ====================================================================
# Copyright (c) 2009 CollabNet.  All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.  The terms
# are also available at http://subversion.tigris.org/license-1.html.
# If newer versions of this license are posted there, you may use a
# newer version instead, at your option.
#
# This software consists of voluntary contributions made by many
# individuals.  For exact contribution history, see the revision
# history and logs, available at http://cvs2svn.tigris.org/.
# ====================================================================

"""Use the generate_blobs.py script to read blobs from RCS files."""

import sys
import os
import subprocess
import cPickle as pickle
import itertools

from cvs2svn_lib.symbol import Trunk
from cvs2svn_lib.util import pjoin
from cvs2svn_lib.cvs_item import CVSRevisionDelete
from cvs2svn_lib.cvs_item import CVSSymbol
from cvs2svn_lib.revision_manager import RevisionRecorder
from cvs2svn_lib.key_generator import KeyGenerator


class ExternalBlobGenerator(RevisionRecorder):
  """Have generate_blobs.py output file revisions to a blob file."""

  def __init__(self, blob_filename):
    self.blob_filename = blob_filename

  def start(self):
    self._mark_generator = KeyGenerator()
    self._popen = subprocess.Popen(
        [
            sys.executable, '-u',
            pjoin(os.path.dirname(__file__), 'generate_blobs.py'),
            self.blob_filename,
            ],
        stdin=subprocess.PIPE,
        )

  def start_file(self, cvs_file_items):
    self._cvs_file_items = cvs_file_items
    self.rcsfilename = cvs_file_items.cvs_file.filename
    self.marks = {}

  def _get_original_source(self, cvs_rev):
    """Return the original source of the contents of CVS_REV.

    Return the first non-delete CVSRevision with the same contents as
    CVS_REV.  'First' here refers to deltatext order; i.e., the very
    first revision is HEAD on trunk, then backwards to the root of a
    branch, then out to the tip of a branch.

    The candidates are all revisions along the CVS delta-dependency
    chain until the next one that has a deltatext (inclusive).  Of the
    candidates, CVSRevisionDeletes are disqualified because, even
    though CVS records their contents, it is impossible to extract
    their fulltext using commands like 'cvs checkout -p'.

    If there is no other CVSRevision that has the same content, return
    CVS_REV itself."""

    # Keep track of the "best" source CVSRevision found so far:
    best_source_rev = None

    for cvs_rev in itertools.chain(
          [cvs_rev], self._cvs_file_items.iter_deltatext_ancestors(cvs_rev)
          ):
      if not isinstance(cvs_rev, CVSRevisionDelete):
        best_source_rev = cvs_rev

      if cvs_rev.deltatext_exists:
        break

    return best_source_rev

  def record_text(self, cvs_rev, log, text):
    if isinstance(cvs_rev, CVSRevisionDelete):
      # There is no need to record a delete revision, and its token
      # will never be needed:
      return None

    source = self._get_original_source(cvs_rev)

    if source.id == cvs_rev.id:
      # Revision is its own source; write it out:
      mark = self._mark_generator.gen_id()
      self.marks[cvs_rev.rev] = mark
      return mark
    else:
      # Return as revision_recorder_token the CVSRevision.id of the
      # original source revision:
      return source.revision_recorder_token

  def finish_file(self, cvs_file_items):
    pickle.dump((self.rcsfilename, self.marks), self._popen.stdin, -1)
    self._popen.stdin.flush()

    # Determine the original source of each CVSSymbol, and store it as
    # the symbol's revision_recorder_token.
    for cvs_item in cvs_file_items.values():
      if isinstance(cvs_item, CVSSymbol):
        cvs_source = cvs_item.get_cvs_revision_source(cvs_file_items)
        cvs_item.revision_recorder_token = cvs_source.revision_recorder_token

    del self._cvs_file_items

  def finish(self):
    self._popen.stdin.close()
    sys.stdout.write('Waiting for generate_blobs.py...')
    returncode = self._popen.wait()
    sys.stdout.write('done (returncode=%s).\n' % (returncode,))


